package main

import (
	"log"
	"sort"

	restful "github.com/emicklei/go-restful"

	"novit.tech/direktil/pkg/localconfig"
)

var clusterSecretKVs = []string{}

func newClusterSecretKV[T any](name string) KVSecrets[T] {
    clusterSecretKVs = append(clusterSecretKVs, name)
    return KVSecrets[T]{"clusters/"+name}
}

func wsListClusters(req *restful.Request, resp *restful.Response) {
	cfg := wsReadConfig(resp)
	if cfg == nil {
		return
	}

	clusterNames := make([]string, len(cfg.Clusters))
	for i, cluster := range cfg.Clusters {
		clusterNames[i] = cluster.Name
	}

	resp.WriteEntity(clusterNames)
}

func wsReadCluster(req *restful.Request, resp *restful.Response) (cluster *localconfig.Cluster) {
	clusterName := req.PathParameter("cluster-name")

	cfg := wsReadConfig(resp)
	if cfg == nil {
		return
	}

	cluster = cfg.Cluster(clusterName)
	if cluster == nil {
		wsNotFound(req, resp)
		return
	}

	return
}

func wsCluster(req *restful.Request, resp *restful.Response) {
	cluster := wsReadCluster(req, resp)
	if cluster == nil {
		return
	}

	resp.WriteEntity(cluster)
}

func wsClusterAddons(req *restful.Request, resp *restful.Response) {
	cluster := wsReadCluster(req, resp)
	if cluster == nil {
		return
	}

	if len(cluster.Addons) == 0 {
		log.Printf("cluster %q has no addons defined", cluster.Name)
		wsNotFound(req, resp)
		return
	}

	cfg := wsReadConfig(resp)
	if cfg == nil {
		return
	}

    sslCfg, err := sslConfigFromLocalConfig(cfg)
    if err != nil {
        wsError(resp, err)
        return
    }

	wsRender(resp, sslCfg, cluster.Addons, cluster)
}

func wsClusterCACert(req *restful.Request, resp *restful.Response) {
	cs := secretData.clusters[req.PathParameter("cluster-name")]
	if cs == nil {
		wsNotFound(req, resp)
		return
	}

	ca := cs.CAs[req.PathParameter("ca-name")]
	if ca == nil {
		wsNotFound(req, resp)
		return
	}

	resp.Write(ca.Cert)
}

func wsClusterSignedCert(req *restful.Request, resp *restful.Response) {
	cs := secretData.clusters[req.PathParameter("cluster-name")]
	if cs == nil {
		wsNotFound(req, resp)
		return
	}

	ca := cs.CAs[req.PathParameter("ca-name")]
	if ca == nil {
		wsNotFound(req, resp)
		return
	}

	name := req.QueryParameter("name")

	if name == "" {
		keys := make([]string, 0, len(ca.Signed))
		for k := range ca.Signed {
			keys = append(keys, k)
		}

		sort.Strings(keys)

		resp.WriteJson(keys, restful.MIME_JSON)
		return
	}

	kc := ca.Signed[name]
	if kc == nil {
		wsNotFound(req, resp)
		return
	}

	resp.Write(kc.Cert)
}
